# coding=utf8

import sys
import traceback
import json

from context import get_json_value
from datetime import timedelta, datetime as dt

DATETIME_FORMAT = '%Y-%m-%dT%TZ'


def format_datetime(d):
    return d.strftime(DATETIME_FORMAT)


def dumpjs(js, ind=4):
    return json.dumps(js, indent=ind)


def print_response(resp):
    if isinstance(resp, dict):
        return dumpjs(resp)
    if isinstance(resp, list):
        return dumpjs(resp)
    return str(resp)


def print_traceback(logger, e, _raise=False):
    exc_info = sys.exc_info()
    stack = traceback.extract_stack()
    tb = traceback.extract_tb(exc_info[2])
    full_tb = stack[:-1] + tb
    exc_line = traceback.format_exception_only(*exc_info[:2])
    traceback_array = [
        'Exception: {}'.format(repr(e)),
        traceback.format_list(full_tb) + exc_line
    ]
    if logger is not None:
        logger.debug('%s\n%s' % (
            traceback_array[0],
            ''.join(traceback_array[1])))
    if _raise:
        raise e

    return traceback_array


def get_next_run(hour, days_pause=0):
    _now = dt.utcnow()
    _next = _now.replace(
        hour=hour,
        minute=0,
        second=0,
        microsecond=0,
        tzinfo=_now.tzinfo
    )
    if _now >= _next:
        _next = _next + timedelta(days=1 + days_pause)
    return _next


def is_string_set(s):
    if not (isinstance(s, str) or isinstance(s, bytes)):
        return False
    if len(s) < 1:
        return False
    return True


def get_string_from_server_config(config, key, default_value=None):
    value = get_json_value(config, key)
    if value is None:
        return default_value
    return str(value)


def get_int_from_server_config(config, key, default_value=None):
    try:
        return int(get_string_from_server_config(config, key))
    except:
        return default_value


def get_bool_from_server_config(config, key, default_value=None):
    value = get_string_from_server_config(config, key)
    if value is None:
        return default_value
    return value == 'true'


def get_object_from_server_config(config, key, default_value=None):
    value = get_json_value(config, key)
    if not isinstance(value, dict):
        return default_value
    return value


def get_list_from_server_config(config, key, default_value=None):
    value = get_json_value(config, key)
    if not isinstance(value, list):
        return default_value
    return value


def update_baseconfig_value(cursor, logger, _class, _key, _parameter, value_type, value):
    try:
        # get ez_value:id
        statement = """\
            SELECT "ez_value:id"
            FROM ez_base_config
            WHERE class='{0}' AND key='{1}' AND parameter='{2}'
        """.format(_class, _key, _parameter)
        cursor.execute(statement)

        res = cursor.fetchall()
        if len(res) != 1:
            return False
        ez_value_id = int(res[0]['ez_value:id'])

        # set value for ez_value:id
        statement = """\
            UPDATE ez_value
            SET {0} = {1}
            WHERE "ez_value:id" = {2}
        """.format(value_type, value, ez_value_id)
        cursor.execute(statement)

        return True
    except Exception as e:
        print_traceback(logger, e)

    return False


class AssetStatus(object):

    def __init__(self, easydb_obj):
        self.easydb_obj = easydb_obj
        self.cloudsight_token = None

        # asset information
        self.eas_id = None
        self.base64_data = None
        self.img_type = None
        self.original_filepath = None

        # status of asset/object
        self.uploaded = False
        self.finished = False
        self.updated = False
        self.failed = False
        self.reason = None

        # ai service results
        self.ai_data_added = False
        self.subject = None
        self.keywords = None

        # instance of CloudsightKeywordCollector
        self.collector = None

    def to_dict(self):
        d = {}
        data = {
            'easydb_obj': self.easydb_obj,
            'cloudsight_token': self.cloudsight_token,
            'eas_id': self.eas_id,
            'img_type': self.img_type,
            'original_filepath': self.original_filepath,
            'uploaded': self.uploaded,
            'finished': self.finished,
            'failed': self.failed,
            'reason': self.reason,
            'subject': self.subject,
            'keywords': self.keywords
        }

        for k in data:
            if data[k] is None:
                continue
            d[k] = data[k]

        return d


class AssetSkippedException(Exception):

    def __init__(self, reason):
        self.reason = reason
        self.message = 'asset skipped, reason: %s' % (reason)
        super(Exception, self).__init__(self.message)


class AssetInfoFailedException(Exception):

    def __init__(self, detail):
        self.detail = detail
        self.message = 'asset info could not be retrieved from object: %s' % (
            detail)
        super(Exception, self).__init__(self.message)
